# SPDX-FileCopyrightText: 2024-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
print("ESP-IDF Kconfig tool")
msg = "Please select a tool to run with command:"
print(
    f"{msg}"
    f"\n{' ' * int(len(msg) / 2)}"
    f"Kconfig file checker. {' ' * 8} (python -m kconfcheck)"
    f"\n{' ' * int(len(msg) / 2)}"
    "Run JSON configuration server. (idf.py confserver or python -m kconfserver)"
    f"\n{' ' * int(len(msg) / 2)}"
    f"Config Generation Tool. {' ' * 6} (python -m kconfgen)"
)



#!/usr/bin/env python
#
# Command line tool to take in ESP-IDF sdkconfig files with project
# settings and output data in multiple formats (update config, generate
# header file, generate .cmake include file, documentation, etc).
#
# Used internally by the ESP-IDF build system. But designed to be
# non-IDF-specific.
#
# SPDX-FileCopyrightText: 2018-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import argparse
import json
import os.path
import re
import sys
import tempfile
import textwrap
from collections import OrderedDict
from collections import defaultdict
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import esp_idf_kconfig.gen_kconfig_doc as gen_kconfig_doc
import esp_kconfiglib.core as kconfiglib
from esp_idf_kconfig import __version__
from esp_kconfiglib.constants import DEP_OP_BEGIN
from esp_kconfiglib.constants import DEP_OP_END
from esp_kconfiglib.constants import SDKCONFIG_DEFAULT_PRAGMA


class DeprecatedOptions(object):
    _RENAME_FILE_NAME = "sdkconfig.rename"
    _RE_DEP_OP_BEGIN = re.compile(DEP_OP_BEGIN)
    _RE_DEP_OP_END = re.compile(DEP_OP_END)

    def __init__(self, config_prefix: str, path_rename_files: List[str] = [], encoding: str = "utf-8"):
        self.config_prefix = config_prefix
        self.encoding = encoding
        # sdkconfig.renames specification: each line contains a pair of config options separated by whitespace(s).
        # The first option is the deprecated one, the second one is the new one.
        # The new option can be prefixed with '!' to indicate inversion (n/not set -> y, y -> n).
        self.rename_line_regex = re.compile(
            rf"#.*|\s*\n|(?P<old>{self.config_prefix}[a-zA-Z_0-9]+)\s+(?P<new>!?{self.config_prefix}[A-Z_0-9]+)"
        )

        # r_dic maps deprecated options to new options; rev_r_dic maps in the opposite direction
        # inversion is a list of deprecated options which will be inverted (n/not set -> y, y -> n)
        self.r_dic: Dict[str, str]
        self.rev_r_dic: DefaultDict[str, List]
        self.inversions: List
        self.r_dic, self.rev_r_dic, self.inversions = self._parse_replacements(path_rename_files)

        # note the '=' at the end of regex for not getting partial match of configs.
        # Also match if the config option is followed by a whitespace, this is the case
        # in sdkconfig.defaults files containing "# CONFIG_MMM_NNN is not set".
        self._RE_CONFIG = re.compile(rf"{self.config_prefix}(\w+)(=|\s+)")

    def parse_line(self, line: str) -> Optional[re.Match]:
        return self.rename_line_regex.match(line)

    def remove_config_prefix(self, string: str) -> str:
        if string.startswith(self.config_prefix):
            return string[len(self.config_prefix) :]
        elif string.startswith("!" + self.config_prefix):
            return string[len("!" + self.config_prefix) :]
        else:
            return ""

    def _parse_replacements(
        self, rename_paths: List[str]
    ) -> Tuple[Dict[str, str], DefaultDict[str, List[str]], List[str]]:
        rep_dic: Dict[str, str] = {}
        rev_rep_dic: DefaultDict[str, List[str]] = defaultdict(list)
        inversions: List[str] = []

        for rename_path in rename_paths:
            with open(rename_path, encoding=self.encoding) as rename_file:
                for line_number, line in enumerate(rename_file, start=1):
                    parsed_line = self.parse_line(line)
                    if not parsed_line:
                        raise RuntimeError(f"Syntax error in {rename_path} (line {line_number})")
                    if not parsed_line["old"]:
                        # empty line or comment
                        continue
                    if parsed_line["old"] in rep_dic:
                        raise RuntimeError(
                            "Error in {} (line {}): Replacement {} exist for {} and new "
                            "replacement {} is defined".format(
                                rename_path,
                                line_number,
                                rep_dic[parsed_line["old"]],
                                parsed_line["old"],
                                parsed_line["new"],
                            )
                        )

                    (dep_opt, new_opt) = (
                        self.remove_config_prefix(parsed_line["old"]),
                        self.remove_config_prefix(parsed_line["new"]),
                    )
                    for opt in (dep_opt, new_opt):
                        if not opt:
                            raise RuntimeError(
                                f"Error in {rename_path} (line {line_number}): "
                                f"Config {opt} is not prefixed with {self.config_prefix}"
                            )
                    if dep_opt == new_opt:
                        raise RuntimeError(
                            f"Error in {rename_path} (line {line_number}): "
                            f"Replacement name is the same as original name ({dep_opt})."
                        )
                    rep_dic[dep_opt] = new_opt
                    rev_rep_dic[new_opt].append(dep_opt)
                    if parsed_line["new"].startswith("!"):
                        inversions.append(dep_opt)

        return rep_dic, rev_rep_dic, inversions

    def is_inversion(self, deprecated_option: str) -> bool:
        return deprecated_option in self.inversions

    def get_deprecated_option(self, new_option: str) -> list:
        return self.rev_r_dic.get(new_option, [])

    def get_new_option(self, deprecated_option: str) -> Optional[str]:
        return self.r_dic.get(deprecated_option, None)

    def replace(self, sdkconfig_in: str, sdkconfig_out: str) -> None:
        replace_enabled = True
        with open(sdkconfig_in, "r", encoding=self.encoding) as input_file, open(
            sdkconfig_out, "w", encoding=self.encoding
        ) as output_file:
            for line_number, line in enumerate(input_file, start=1):
                if self._RE_DEP_OP_BEGIN.search(line):  # Begin of deprecated options
                    replace_enabled = False
                elif self._RE_DEP_OP_END.search(line):  # End of deprecated options
                    replace_enabled = True
                elif replace_enabled:
                    m = self._RE_CONFIG.search(line)
                    if m and m.group(1) in self.r_dic:  # Deprecated option found
                        depr_opt = self.config_prefix + m.group(1)
                        new_opt = self.config_prefix + self.r_dic[m.group(1)]
                        line = self.replace_line(line, depr_opt=depr_opt, new_opt=new_opt)
                        print(
                            "{}:{} {} was replaced with {} {}".format(
                                sdkconfig_in,
                                line_number,
                                depr_opt,
                                new_opt,
                                "and inverted" if depr_opt[len(self.config_prefix) :] in self.inversions else "",
                            )
                        )
                output_file.write(line)

    def replace_line(self, line: str, depr_opt: str, new_opt: str, depr_to_new: bool = True) -> str:
        depr_name = self.remove_config_prefix(depr_opt)
        to_replace = depr_opt if depr_to_new else new_opt
        replace_with = new_opt if depr_to_new else depr_opt

        # NOTE: We want deprecated options to be treated as user-set: they are not part
        #       of the normal menu tree/menuconfig TUI etc. and even direct change in sdkconfig
        #       will take no effect. For that reason, skipping the default value assignment
        #       machinery makes the process easier without any drawbacks.
        if line.startswith(SDKCONFIG_DEFAULT_PRAGMA):
            line = line[len(SDKCONFIG_DEFAULT_PRAGMA) + 1 :]

        if depr_name in self.inversions:
            if any(substring in line for substring in ("is not set", "=n")):
                line = f"{replace_with}=y\n"
            else:
                line = f"{replace_with}=n\n"
        else:
            line = line.replace(to_replace, replace_with)
        return line

    def append_doc(
        self,
        config: kconfiglib.Kconfig,
        visibility: gen_kconfig_doc.ConfigTargetVisibility,
        path_output: str,
    ) -> None:
        def option_was_written(opt: str) -> bool:
            # named choices were written if any of the symbols in the choice were visible
            if new_opt in config.named_choices:
                syms = config.named_choices[new_opt].syms
                for s in syms:
                    if any(visibility.visible(node) for node in s.nodes):
                        return True
                return False
            else:
                try:
                    # otherwise if any of the nodes associated with the option was visible
                    return any(visibility.visible(node) for node in config.syms[opt].nodes)
                except KeyError:
                    return False

        if len(self.r_dic) > 0:
            with open(path_output, "a", encoding=self.encoding) as f_o:
                header = "Deprecated options and their replacements"
                f_o.write(".. _configuration-deprecated-options:\n\n{}\n{}\n\n".format(header, "-" * len(header)))
                for dep_opt in sorted(self.r_dic):
                    new_opt = self.r_dic[dep_opt]
                    if option_was_written(new_opt) and (
                        new_opt not in config.syms or config.syms[new_opt].choice is None
                    ):
                        # everything except config for a choice (no link reference for those in the docs)
                        f_o.write(
                            "- {}{} ({}:ref:`{}{}`)\n".format(
                                config.config_prefix,
                                dep_opt,
                                "inversion of " if dep_opt in self.inversions else "",
                                config.config_prefix,
                                new_opt,
                            )
                        )

                        if new_opt in config.named_choices:
                            # here are printed config options which were filtered out
                            syms = config.named_choices[new_opt].syms
                            for sym in syms:
                                if sym.name in self.rev_r_dic:
                                    # only if the symbol has been renamed
                                    dep_names = self.rev_r_dic[sym.name]
                                    dep_names = [config.config_prefix + name for name in dep_names]
                                    # config options doesn't have references
                                    f_o.write("    - {}\n".format(", ".join(dep_names)))

    def append_config(self, config: kconfiglib.Kconfig, path_output: str) -> None:
        tmp_list = []

        def append_config_node_process(node: kconfiglib.MenuNode) -> None:
            item = node.item
            if isinstance(item, kconfiglib.Symbol) and item.env_var is None:
                if item.name in self.rev_r_dic:
                    c_string = item.config_string
                    if c_string:
                        for dep_name in self.rev_r_dic[item.name]:
                            tmp_list.append(
                                self.replace_line(
                                    c_string,
                                    depr_opt=self.config_prefix + dep_name,
                                    new_opt=self.config_prefix + item.name,
                                    depr_to_new=False,
                                )
                            )

        for n in config.node_iter():
            append_config_node_process(n)

        if len(tmp_list) > 0:
            with open(path_output, "a", encoding=self.encoding) as f_o:
                f_o.write("\n{}\n".format(DEP_OP_BEGIN))
                f_o.writelines(tmp_list)
                f_o.write("{}\n".format(DEP_OP_END))

    def append_header(self, config: kconfiglib.Kconfig, path_output: str) -> None:
        def _opt_defined(opt):
            if opt.orig_type == kconfiglib.BOOL and opt.str_value != "n":
                opt_defined = True
            elif opt.orig_type in (kconfiglib.INT, kconfiglib.STRING, kconfiglib.HEX) and opt.str_value != "":
                opt_defined = True
            else:
                opt_defined = False
            return opt_defined

        if len(self.r_dic) > 0:
            with open(path_output, "a", encoding=self.encoding) as output_file:
                output_file.write("\n/* List of deprecated options */\n")
                for dep_opt in sorted(self.r_dic):
                    new_opt = self.r_dic[dep_opt]
                    if new_opt in config.syms and _opt_defined(config.syms[new_opt]):
                        output_file.write(
                            "#define {}{} {}{}{}\n".format(
                                self.config_prefix,
                                dep_opt,
                                "!" if dep_opt in self.inversions else "",
                                self.config_prefix,
                                new_opt,
                            )
                        )


def write_config(deprecated_options: DeprecatedOptions, config: kconfiglib.Kconfig, filename: str) -> None:
    idf_version = os.environ.get("IDF_VERSION", "")
    CONFIG_HEADING = textwrap.dedent(
        f"""\
    #
    # Automatically generated file. DO NOT EDIT.
    # Espressif IoT Development Framework (ESP-IDF) {idf_version} Project Configuration
    #
    """
    )
    config.write_config(filename, header=CONFIG_HEADING)
    deprecated_options.append_config(config, filename)


def min_config_with_labels(config: kconfiglib.Kconfig, header: str) -> str:
    """Return minimal config containing menu labels"""
    all_options = config._config_contents("").splitlines()
    min_options = config._min_config_contents("").splitlines()

    end_regex = re.compile(r"^# end of (.*)$").match
    # `label_path` marks the current path from the root; labels represent tree nodes.
    # The path is represented as an ordered Dict; the last element in the Dict is closest to the leaves.
    # The label name is the key, and a boolean value indicates whether the label was added to the output
    # (i.e., the label is used in the output).
    # We need to track used labels to ensure correct timing when printing the label ending.
    # Note that label names are not necessarily unique, so the order is significant.
    label_path: OrderedDict[str, bool] = OrderedDict()
    output = [header]
    possibly_label = False
    current = None  # None stands for tree root
    comments = []
    config_has_default_value = False

    # Using depth search first, we go down the tree and save the path from the root.
    # If we find an option from min config, we update the whole path to used (True) and print all menu labels.
    # When we go back up the tree, we print all label endings that were marked as used
    for line in all_options:
        end_match = end_regex(line)
        if end_match:
            # we have found an end of a menu section
            current = end_match.group(1)
            label = None
            while label != current and label_path:
                # remove any labels that appeared outside of used sections
                label, used = label_path.popitem()
                if used and label != current:
                    # used label without end -> comment
                    comments.append(label)
            if used:
                # menu label was used - print menu label ending
                output.append(f"{line}\n")
            # find the new current (the last in list); if tree is empty return back to the root
            current = next(reversed(label_path.keys())) if label_path else None
        elif line == "#":
            # starting/ending possible menu label
            possibly_label = not possibly_label
        elif possibly_label:
            current = line[2:]  # remove leading '# '
            label_path[current] = False  # label not yet used
        elif line in min_options:
            # minimal config option detected
            for label, used in label_path.items():
                # print all menu labels that were not printed yet
                if not used:
                    output.append(f"\n#\n# {label}\n#\n")
                # mark the whole path from root as 'used'
                label_path[label] = True

            if line.strip() == config.comment_default_value:  # indicator that the next config has a default value
                config_has_default_value = True
            else:
                comment_value_is_default = f"{config.comment_default_value}\n" if config_has_default_value else ""
                output.append(f"{comment_value_is_default}{line}\n")
                config_has_default_value = False
        else:
            config_has_default_value = False
    # Remove comments from minimal config, while keeping menu labels
    for comment in comments:
        output.remove(f"\n#\n# {comment}\n#\n")
    return "".join(output)


def write_min_config(_: DeprecatedOptions, config: kconfiglib.Kconfig, filename: str) -> None:
    idf_version = os.environ.get("IDF_VERSION", "")
    target_symbol = config.syms["IDF_TARGET"]
    # 'esp32` is hardcoded here because the default value of IDF_TARGET is set on the first run from the environment
    # variable. I.E. `esp32  is not defined as default value.
    write_target = target_symbol.str_value != "esp32"

    CONFIG_HEADING = textwrap.dedent(
        f"""\
    # This file was generated using idf.py save-defconfig. It can be edited manually.
    # Espressif IoT Development Framework (ESP-IDF) {idf_version} Project Minimal Configuration
    #
    {target_symbol.config_string if write_target else ""}\
    """
    )

    if os.environ.get("ESP_IDF_KCONFIG_MIN_LABELS", False) == "1":
        lines = min_config_with_labels(config, CONFIG_HEADING).splitlines()
    else:
        lines = config._min_config_contents(header=CONFIG_HEADING).splitlines()

    # convert `# CONFIG_XY is not set` to `CONFIG_XY=n` to improve readability
    unset_match = re.compile(r"# {}([^ ]+) is not set".format(config.config_prefix)).match
    for idx, line in enumerate(lines):
        match = unset_match(line)
        if match:
            lines[idx] = f"{config.config_prefix}{match.group(1)}=n"
    lines[-1] += "\n"
    config._write_if_changed(filename, "\n".join(lines))


def write_header(deprecated_options: DeprecatedOptions, config: kconfiglib.Kconfig, filename: str) -> None:
    idf_version = os.environ.get("IDF_VERSION", "")
    CONFIG_HEADING = f"""/*
 * Automatically generated file. DO NOT EDIT.
 * Espressif IoT Development Framework (ESP-IDF) {idf_version} Configuration Header
 */
#pragma once
"""
    config.write_autoconf(filename, header=CONFIG_HEADING)
    deprecated_options.append_header(config, filename)


def write_cmake(deprecated_options: DeprecatedOptions, config: kconfiglib.Kconfig, filename: str) -> None:
    with open(filename, "w", encoding=config._encoding) as f:
        tmp_dep_list = []
        prefix = config.config_prefix

        f.write(
            textwrap.dedent(
                """#
                # Automatically generated file. DO NOT EDIT.
                # Espressif IoT Development Framework (ESP-IDF) Configuration cmake include file
                #
                """
            )
        )

        configs_list = list()

        def write_node(node: kconfiglib.MenuNode) -> None:
            sym = node.item
            if not isinstance(sym, kconfiglib.Symbol):
                return

            if sym.config_string:
                val = sym.str_value
                if sym.orig_type == kconfiglib.BOOL and val == "n":
                    val = ""  # write unset values as empty variables
                elif sym.orig_type == kconfiglib.STRING:
                    val = kconfiglib.escape(val)
                elif sym.orig_type == kconfiglib.HEX:
                    val = hex(int(val, 16))  # ensure 0x prefix
                f.write(f'set({prefix}{sym.name} "{val}")\n')

                configs_list.append(prefix + sym.name)
                dep_opts = deprecated_options.get_deprecated_option(sym.name)
                for opt in dep_opts:
                    if deprecated_options.is_inversion(opt) and sym.orig_type == kconfiglib.BOOL:
                        val = "y" if not val else ""
                    tmp_dep_list.append('set({}{} "{}")\n'.format(prefix, opt, val))
                    configs_list.append(prefix + opt)

        for n in config.node_iter():
            write_node(n)
        f.write("set(CONFIGS_LIST {})".format(";".join(configs_list)))

        if len(tmp_dep_list) > 0:
            f.write("\n# List of deprecated options for backward compatibility\n")
            f.writelines(tmp_dep_list)


def get_json_values(config: kconfiglib.Kconfig) -> dict:
    config_dict = {}

    def write_node(node: kconfiglib.MenuNode) -> None:
        sym = node.item
        if not isinstance(sym, kconfiglib.Symbol):
            return

        if sym.config_string:
            candidate_val: str = sym.str_value
            if not candidate_val and sym.type in (
                kconfiglib.INT,
                kconfiglib.HEX,
            ):
                print(
                    f"warning: {sym.name} has no value set in the configuration."
                    " This can be caused e.g. by missing default value for the current chip version."
                )
                val: Optional[Union[str, bool, int]] = None
            elif sym.type == kconfiglib.BOOL:
                val = candidate_val != "n"
            elif sym.type == kconfiglib.HEX:
                val = int(candidate_val, 16)
            elif sym.type == kconfiglib.INT:
                val = int(candidate_val)
            else:
                val = candidate_val
            config_dict[sym.name] = val

    for n in config.node_iter(False):
        write_node(n)
    return config_dict


def write_json(_: DeprecatedOptions, config: kconfiglib.Kconfig, filename: str) -> None:
    config_dict = get_json_values(config)
    with open(filename, "w", encoding=config._encoding) as f:
        json.dump(config_dict, f, indent=4, sort_keys=True)


def write_json_menus(_: DeprecatedOptions, config: kconfiglib.Kconfig, filename: str) -> None:
    existing_ids: Set[str] = set()
    result: List = []  # root level items
    node_lookup: Dict = {}  # lookup from MenuNode to an item in result

    def write_node(node: kconfiglib.MenuNode) -> None:
        try:
            json_parent = node_lookup[node.parent]["children"]
        except KeyError:
            assert (
                node.parent not in node_lookup
            )  # if fails, we have a parent node with no "children" entity (ie a bug)
            json_parent = result  # root level node

        # node.kconfig.y means node has no dependency,
        if node.dep is node.kconfig.y:
            depends = None
        else:
            depends = kconfiglib.expr_str(node.dep)

        try:
            # node.is_menuconfig is True in newer kconfiglibs for menus and choices as well
            is_menuconfig = node.is_menuconfig and isinstance(node.item, kconfiglib.Symbol)
        except AttributeError:
            is_menuconfig = False

        new_json: Dict[str, Any] = {}
        if node.item == kconfiglib.MENU or is_menuconfig:
            new_json = {
                "type": "menu",
                "title": node.prompt[0] if node.prompt else "",
                "depends_on": depends,
                "children": [],
            }
            if is_menuconfig:
                sym = node.item
                new_json["name"] = sym.name  # type: ignore
                new_json["help"] = node.help
                new_json["is_menuconfig"] = is_menuconfig
                greatest_range = None
                if isinstance(sym, (kconfiglib.Symbol, kconfiglib.MenuNode)) and len(sym.ranges) > 0:
                    # Note: Evaluating the condition using kconfiglib's expr_value
                    # should have one condition which is true
                    for min_range, max_range, cond_expr in sym.ranges:
                        if kconfiglib.expr_value(cond_expr):
                            greatest_range = [min_range, max_range]
                new_json["range"] = greatest_range

        elif isinstance(node.item, kconfiglib.Symbol):
            sym = node.item
            greatest_range = None
            if len(sym.ranges) > 0:
                # Note: Evaluating the condition using kconfiglib's expr_value
                # should have one condition which is true
                for min_range, max_range, cond_expr in sym.ranges:
                    if kconfiglib.expr_value(cond_expr):
                        base = 16 if sym.type == kconfiglib.HEX else 10
                        greatest_range = [
                            int(min_range.str_value, base),
                            int(max_range.str_value, base),
                        ]
                        break

            new_json = {
                "type": kconfiglib.TYPE_TO_STR[sym.type],
                "name": sym.name,
                "title": node.prompt[0] if node.prompt else None,
                "depends_on": depends,
                "help": node.help,
                "range": greatest_range,
                "children": [],
            }
        elif isinstance(node.item, kconfiglib.Choice):
            choice = node.item
            new_json = {
                "type": "choice",
                "title": node.prompt[0] if node.prompt else "",
                "name": choice.name,
                "depends_on": depends,
                "help": node.help,
                "children": [],
            }

        if new_json:
            if node.id in existing_ids:
                raise RuntimeError(
                    "Config file contains two items with the same id: "
                    f" {node.id} ({node.prompt[0] if node.prompt else ''}). "
                    "Please rename one of these items to avoid ambiguity."
                )
            new_json["id"] = node.id

            json_parent.append(new_json)
            node_lookup[node] = new_json

    for n in config.node_iter():
        write_node(n)
    with open(filename, "w", encoding=config._encoding) as f:
        f.write(json.dumps(result, sort_keys=True, indent=4))


def write_docs(deprecated_options: DeprecatedOptions, config: kconfiglib.Kconfig, filename: str) -> None:
    try:
        target = os.environ["IDF_TARGET"]
    except KeyError:
        print("IDF_TARGET environment variable must be defined!")
        sys.exit(1)

    visibility = gen_kconfig_doc.ConfigTargetVisibility(config, target)
    gen_kconfig_doc.write_docs(config, visibility, filename)
    deprecated_options.append_doc(config, visibility, filename)


def write_report(_: DeprecatedOptions, config: kconfiglib.Kconfig, filename: str) -> None:
    config.report.output_json(filename)


def update_if_changed(source: str, destination: str, encoding: str) -> None:
    with open(source, "r", encoding=encoding) as f:
        source_contents = f.read()

    if os.path.exists(destination):
        with open(destination, "r", encoding=encoding) as f:
            dest_contents = f.read()
        if source_contents == dest_contents:
            return  # nothing to update

    with open(destination, "w", encoding=encoding) as f:
        f.write(source_contents)


OUTPUT_FORMATS = {
    "config": write_config,
    "header": write_header,
    "cmake": write_cmake,
    "docs": write_docs,
    "json": write_json,
    "json_menus": write_json_menus,
    "savedefconfig": write_min_config,
    "report": write_report,
}


def main():
    parser = argparse.ArgumentParser(
        description="kconfgen.py v%s - Config Generation Tool" % __version__,
        prog=os.path.basename(sys.argv[0]),
    )

    parser.add_argument("--config", help="Project configuration settings", nargs="?", default=None)

    parser.add_argument(
        "--defaults",
        help="Optional project defaults file, used if --config file doesn't exist. "
        "Multiple files can be specified using multiple --defaults arguments.",
        nargs="?",
        default=[],
        action="append",
    )

    parser.add_argument("--kconfig", help="Kconfig file with config item definitions", required=True)

    parser.add_argument(
        "--sdkconfig-rename",
        help="File with deprecated Kconfig options",
        required=False,
    )

    parser.add_argument(
        "--dont-write-deprecated",
        help="Do not write compatibility statements for deprecated values",
        action="store_true",
    )

    parser.add_argument(
        "--output",
        nargs=2,
        action="append",
        help="Write output file (format and output filename)",
        metavar=("FORMAT", "FILENAME"),
        default=[],
    )

    parser.add_argument(
        "--env",
        action="append",
        default=[],
        help="Environment to set when evaluating the config file",
        metavar="NAME=VAL",
    )

    parser.add_argument(
        "--env-file",
        type=argparse.FileType("r"),
        help="Optional file to load environment variables from. Contents "
        "should be a JSON object where each key/value pair is a variable.",
    )

    parser.add_argument(
        "--list-separator",
        choices=["space", "semicolon"],
        default="space",
        help="Separator used in environment list variables (COMPONENT_SDKCONFIG_RENAMES)",
    )
    args = parser.parse_args()

    for fmt, filename in args.output:
        if fmt not in OUTPUT_FORMATS.keys():
            print("Format '%s' not recognised. Known formats: %s" % (fmt, OUTPUT_FORMATS.keys()))
            sys.exit(1)

    try:
        args.env = [(name, value) for (name, value) in (e.split("=", 1) for e in args.env)]
    except ValueError:
        print("--env arguments must each contain =. To unset an environment variable, use 'ENV='")
        sys.exit(1)

    for name, value in args.env:
        os.environ[name] = value

    if args.env_file is not None:
        env = json.load(args.env_file)
        os.environ.update(env)
    parser_version = int(os.environ.get("KCONFIG_PARSER_VERSION", "1"))
    # TODO Once ESP-IDF will fully support kconfig report, we should switch to "quiet" as default
    #      to avoid printing the report several times during the build.
    print_report = os.environ.get("KCONFIG_REPORT_VERBOSITY", "default") != "quiet"
    config = kconfiglib.Kconfig(
        args.kconfig,
        parser_version=parser_version,
        print_report=(
            print_report
            and not (
                args.config and os.path.exists(args.config)
            )  # if sdkconfig file exists, we'll report after it is loaded
            and len(args.defaults) == 0  # if defaults are loaded, report will be printed after that
        ),
    )
    kconfig_encoding = config._encoding

    sdkconfig_renames_sep = ";" if args.list_separator == "semicolon" else " "
    sdkconfig_renames = [args.sdkconfig_rename] if args.sdkconfig_rename else []
    sdkconfig_renames_from_env = os.environ.get("COMPONENT_SDKCONFIG_RENAMES")
    if sdkconfig_renames_from_env:
        sdkconfig_renames += sdkconfig_renames_from_env.split(sdkconfig_renames_sep)
    deprecated_options = DeprecatedOptions(
        config.config_prefix, path_rename_files=sdkconfig_renames, encoding=kconfig_encoding
    )

    if len(args.defaults) > 0:

        def _replace_empty_assignments(path_in, path_out):  # empty assignment: CONFIG_FOO=
            with open(path_in, "r", encoding=kconfig_encoding) as f_in, open(
                path_out, "w", encoding=kconfig_encoding
            ) as f_out:
                for line_num, line in enumerate(f_in, start=1):
                    line = line.strip()
                    if line.endswith("="):
                        line += "n"
                        print("{}:{} line was updated to {}".format(path_out, line_num, line))
                    f_out.write(line)
                    f_out.write("\n")

        # always load defaults first, so any items which are not defined in the args.config
        # will have the default defined in the defaults file
        for name in args.defaults:
            print("Loading defaults file %s..." % name)
            if not os.path.exists(name):
                raise RuntimeError("Defaults file not found: %s" % name)
            try:
                with tempfile.NamedTemporaryFile(
                    prefix="kconfgen_tmp", mode="w+", delete=False, encoding=kconfig_encoding
                ) as f:
                    temp_file1 = f.name
                with tempfile.NamedTemporaryFile(
                    prefix="kconfgen_tmp", mode="w+", delete=False, encoding=kconfig_encoding
                ) as f:
                    temp_file2 = f.name
                deprecated_options.replace(sdkconfig_in=name, sdkconfig_out=temp_file1)
                _replace_empty_assignments(temp_file1, temp_file2)
                config.load_config(temp_file2, replace=False)

                for symbol, value in config.missing_syms:
                    if deprecated_options.get_new_option(symbol) is None:
                        print(f"warning: unknown kconfig symbol '{symbol}' assigned to '{value}' in {name}")
            finally:
                try:
                    os.remove(temp_file1)
                    os.remove(temp_file2)
                except OSError:
                    pass
        if print_report and not (args.config and os.path.exists(args.config)):
            # if no sdkconfig file exists, print the report now
            config.report.print_report()

    # If previous sdkconfig file exists, load it
    if args.config and os.path.exists(args.config):
        # ... but replace deprecated options before that
        with tempfile.NamedTemporaryFile(
            prefix="kconfgen_tmp", mode="w+", delete=False, encoding=kconfig_encoding
        ) as f:
            temp_file = f.name
        try:
            deprecated_options.replace(sdkconfig_in=args.config, sdkconfig_out=temp_file)
            config.load_config(temp_file, replace=False, print_report=print_report)
            update_if_changed(temp_file, args.config, encoding=kconfig_encoding)
        finally:
            try:
                os.remove(temp_file)
            except OSError:
                pass

    if args.dont_write_deprecated:
        # The deprecated object was useful until now for replacements. Now it will be redefined with no configurations
        # and as the consequence, it won't generate output with deprecated statements.
        deprecated_options = DeprecatedOptions("", path_rename_files=[])

    # Output the files specified in the arguments
    for output_type, filename in args.output:
        with tempfile.NamedTemporaryFile(
            prefix="kconfgen_tmp", mode="w+", delete=False, encoding=kconfig_encoding
        ) as f:
            temp_file = f.name
        try:
            output_function = OUTPUT_FORMATS[output_type]
            output_function(deprecated_options, config, temp_file)
            update_if_changed(temp_file, filename, encoding=kconfig_encoding)
        finally:
            try:
                os.remove(temp_file)
            except OSError:
                pass



# SPDX-FileCopyrightText: 2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
from esp_docs.conf_docs import *  # noqa: F403,F401
from esp_docs.conf_docs import extensions
from esp_docs.conf_docs import html_context

languages = ["en"]
# link roles config
github_repo = "espressif/esp-idf-kconfig"

# context used by sphinx_idf_theme
html_context["github_user"] = "espressif"
html_context["github_repo"] = "esp-idf-kconfig"

html_static_path = ["../_static"]

# Conditional content
extensions += ["esp_docs.esp_extensions.dummy_build_system"]

# Extra options required by sphinx_idf_theme
project_slug = "esp-idf-kconfig"

versions_url = "./_static/esp_idf_kconfig_versions.js"


def conf_setup(app, config):
    config.html_baseurl = f"https://docs.espressif.com/projects/esp-idf-kconfig/{config.language}/stable/"



# SPDX-FileCopyrightText: 2024-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple


def _parse_path(path: str, sep: Optional[str] = None) -> Set:
    ret = set()
    with open(path, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line.startswith("#") and len(line) > 0:
                ret.add(line.split(sep)[0])
    return ret


def _prepare_deprecated_options(
    includes: List[str], exclude_submodules: List[str], files: List[str]
) -> Tuple[List[str], Set[str], Tuple]:
    deprecated_options = set()

    # used as "startswith" pattern to match sdkconfig.[ci|defaults] variants
    files_to_check_pattern = ("sdkconfig.ci", "sdkconfig.defaults")

    # Ignored directories (makes sense only when run on IDF_PATH)
    # Note: ignore_dirs is a tuple in order to be able to use it directly with the startswith() built-in function which
    # accepts tuples but no lists.
    ignore_dirs: Tuple = tuple(exclude_submodules) if exclude_submodules else tuple()

    # sdkconfig.rename files can be explicitly specified in the command line as well,
    # but they are not checked, only used as a source of deprecated options.
    for file in tuple(files):
        if "sdkconfig.rename" in file:
            deprecated_options.update(_parse_path(file))
            files.remove(file)

    # Recursively search for sdkconfig files in the specified directories
    # sdkconfig.rename: used as a source of deprecated options
    # sdkconfig.[ci|defaults]: added to list of files to be checked
    if includes:
        for directory in includes:
            for root, _, filenames in os.walk(directory):
                for filename in filenames:
                    full_path = os.path.join(root, filename)
                    if filename.startswith(files_to_check_pattern) and not full_path.startswith(ignore_dirs):
                        files.append(full_path)
                    elif filename == "sdkconfig.rename":
                        deprecated_options.update(_parse_path(full_path))
                    elif full_path.startswith(ignore_dirs):
                        print(f"{full_path}: Ignored")

    # Ignoring includes/excludes, we still want to gather all the deprecated values.
    # IDF_PATH should be set most of the time, but if not, use the current directory as a fallback.
    # When running kconfcheck as a pre-commit hook in ESP-IDF repo, os.getcwd() will be the same as IDF_PATH.
    project_path = os.environ.get("IDF_PATH", None)
    if not project_path:
        project_path = os.getcwd()
        print(f"kconfcheck: IDF_PATH is not set. Using {project_path} as a fallback.")
    for root, _, filenames in os.walk(project_path):
        for filename in filenames:
            if filename == "sdkconfig.rename":
                deprecated_options.update(_parse_path(os.path.join(root, filename)))

    return files, deprecated_options, ignore_dirs


def check_deprecated_options(
    file_full_path: str,
    deprecated_options: Set[str] = set(),
    ignore_dirs: Tuple = tuple(),
) -> Optional[bool]:
    if file_full_path in ignore_dirs:
        print(f"{file_full_path}: Ignored")
        return None
    used_options = _parse_path(file_full_path, "=")
    used_deprecated_options = deprecated_options.intersection(used_options)
    if len(used_deprecated_options) > 0:
        print(f"{file_full_path}: The following options are deprecated: {', '.join(used_deprecated_options)}")
        return False
    else:
        print(f"{file_full_path}: OK")
        return True



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import sys

from esp_menuconfig.core import _main

if __name__ == "__main__":
    try:
        _main()
    except Exception as e:
        print(f"A fatal error occurred: {e}", file=sys.stderr)
        sys.exit(2)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

"""
New Kconfig Report class to log messages.

Instead of continuously logging messages,KconfigReport class will store
the messages and print them at the end of the parsing process as one report.
"""

import json
import textwrap
from collections import defaultdict
from typing import TYPE_CHECKING
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

from .constants import DefaultsPolicy

if TYPE_CHECKING:
    from .core import Choice
    from .core import Kconfig
    from .core import Symbol

import os
from abc import ABC
from abc import abstractmethod

from rich import print as rprint
from rich.box import HORIZONTALS
from rich.console import Console
from rich.table import Table

PRAGMA_PREFIX = "# ignore:"

STATUS_NONE = 0
STATUS_OK = 1
STATUS_OK_WITH_INFO = 2
STATUS_WARNING = 3
STATUS_ERROR = 4

_INDENT = " " * 4
VERBOSITY_QUIET = "quiet"  # Report only if there is an error
VERBOSITY_DEFAULT = "default"  # Report standard information
VERBOSITY_VERBOSE = "verbose"  # Report everything every time

AREA_TITLE_STYLE = "bold blue"
INFO_STRING_STYLE = "italic"
SUBTITLE_STYLE = "bold"


class Area(ABC):
    """
    Abstract class holding the base structure of every area in the report.
    """

    def __init__(self, title: str, ignore_codes: Set[str], info_string: str):
        """
        title:
        info_string:
            Both are used to describe the area in the report.
            Title is printed every time specific area is printed, info string provides additional information
            about the area, which may further help to understand the issue.
        """
        self.title: str = title
        self.info_string: str = info_string

        """
        ignored_sc:
            Dictionary of Symbol/Choice objects (not only names!) which should be ignored in the report.
        """
        self.ignored_sc: Dict[str, Set[str]] = {"configs": set(), "choices": set()}
        """
        ignore_codes:
            Ignore codes which are used to ignore reporting of certain areas for certain symbols
            (e.g. multiple-definition).
            Ignore codes are written after PRAGMA_PREFIX in the Kconfig files (e.g. # ignore: multiple-definition).
        """
        self.ignore_codes: Set[str] = ignore_codes

    @abstractmethod
    def add_record(self, sym_or_choice: "Union[Symbol, Choice]", **kwargs: Optional[dict]) -> None:
        """
        Adding a new record to the area.
        All records relates to the specific Symbol/Choice and may contain additional information in kwargs.
        """
        pass

    def add_ignore(self, sym_or_choice: "Union[Symbol, Choice]") -> None:
        """
        Add a Symbol/Choice to the ignore list of the area.
        Raises AttributeError if the area does not support ignore codes.
        """
        raise AttributeError(f"{self.__class__.__name__} does not support ignore codes.")

    @abstractmethod
    def report_severity(self) -> int:
        """
        If given area has nothing to report, STATUS_OK should be returned.
        Otherwise, STATUS_OK_WITH_INFO, STATUS_WARNING or STATUS_ERROR should be returned,
        depending how severe record in given area is.
        """
        pass

    @abstractmethod
    def print(self, verbosity: str) -> Optional[Table]:
        """
        Print the area sub-report.
        """
        pass

    @abstractmethod
    def return_json(self) -> Optional[dict]:
        """
        Return the area report in JSON format.
        """
        ret_json = dict()
        ret_json["title"] = self.title
        ret_json["severity"] = self.severity_to_str(self.report_severity())
        return ret_json

    @abstractmethod
    def reset(self) -> None:
        """
        Reset the area to its initial state, clearing all records and data.
        """
        pass

    @staticmethod
    def severity_to_str(severity: int) -> str:
        if severity == STATUS_OK:
            return "OK"
        elif severity == STATUS_OK_WITH_INFO:
            return "Info"
        elif severity == STATUS_WARNING:
            return "Warning"
        else:
            return "Error"


class DefaultValuesArea(Area):
    """
    Area for reporting any problems with default values:
    * Different values between sdkconfig and Kconfig files
    * Promptless default values with different values between sdkconfig and Kconfig files
      This is normally not a problem, but a feature. However, devs may want to know about it.
    """

    def __init__(self, defaults_policy: DefaultsPolicy, verbosity: str):
        super().__init__(
            title="Default Value Mismatch",
            ignore_codes=set(),
            info_string=textwrap.dedent(
                """\
                This area reports issues with default values of the config options.
                For more information, please visit https://docs.espressif.com/projects/esp-idf-kconfig/en/latest/kconfiglib/configuration-report.html#default-value-mismatch
                """
            ),
        )
        self.defaults_policy: DefaultsPolicy = defaults_policy
        self.verbosity: str = verbosity

        self.changed_defaults: Set[Tuple[str, str, str]] = set()
        # Changed configs without prompts should not be reported as it's not something user should care about.
        # However, it may be useful to report them in verbose mode for devs.
        self.changed_values_promptless: Set[Tuple[str, str, str, bool]] = set()
        self.changed_choices: Set[Tuple[str, str, str]] = set()

    def add_record(self, sym_or_choice: "Union[Symbol, Choice]", **kwargs: Optional[dict]) -> None:
        promptless: bool = kwargs.get("promptless", False)  # type: ignore
        record_type: str = kwargs.get("record_type", "symbol")  # type: ignore
        if record_type == "symbol":  # Symbol
            record = (
                str(sym_or_choice.name),
                str(sym_or_choice.str_value),
                str(getattr(sym_or_choice, "_sdkconfig_value", "") or ""),
            )
            if not promptless:
                self.changed_defaults.add(record)
            else:
                # sdkconfig value is still set even for promptless symbols, so we can decide
                # if sdkconfig contained default value or not
                record_with_default_flag = record + (getattr(sym_or_choice, "_user_value", None) is not None,)
                self.changed_values_promptless.add(record_with_default_flag)
        else:  # Choice
            record = (
                str(sym_or_choice.name or "nameless" + sym_or_choice.name_and_loc),
                str(sym_or_choice.selection.name if sym_or_choice.selection else "choice deselected"),  # type: ignore
                str(kwargs.get("sdkconfig_selection", False)),
            )
            self.changed_choices.add(record)

    def report_severity(self) -> int:
        if (
            not self.changed_defaults
            and not (self.changed_values_promptless and self.verbosity == VERBOSITY_VERBOSE)
            and not self.changed_choices
        ):
            return STATUS_OK
        if (
            self.changed_defaults
            or (self.changed_values_promptless and self.verbosity == VERBOSITY_VERBOSE)
            or self.changed_choices
        ):
            return STATUS_OK_WITH_INFO
        else:  # This should not happen, but just in case
            return STATUS_ERROR

    def _nothing_to_report(self, verbosity: str = VERBOSITY_VERBOSE) -> bool:
        """
        Check if there is nothing to report in the area.
        """
        return (
            not self.changed_defaults
            and not self.changed_choices
            and (verbosity != VERBOSITY_VERBOSE or not self.changed_values_promptless)
        )

    def print(self, verbosity: str) -> Optional[Table]:
        # No changed defaults or only promptless changed defaults without verbosity VERBOSITY_VERBOSE
        # -> nothing to report
        if self._nothing_to_report(verbosity):
            return None

        table = Table(title=self.title, title_justify="left", show_header=False, title_style=AREA_TITLE_STYLE)
        table.add_row(self.defaults_policy.description, style=INFO_STRING_STYLE)
        table.box = HORIZONTALS
        table.add_column(
            "",
            justify="left",
            no_wrap=False,
        )
        if verbosity == VERBOSITY_VERBOSE:
            table.add_row(self.info_string, style=INFO_STRING_STYLE)

        if self.changed_defaults:
            table.add_row(
                "Config symbols with different default value between sdkconfig and Kconfig", style=SUBTITLE_STYLE
            )
            for sym_name, kconfig_value, sdkconfig_value in self.changed_defaults:
                table.add_row(
                    f"{sym_name}: Kconfig default value: {kconfig_value}, sdkconfig default value: {sdkconfig_value}"
                )
            table.add_row("")

        if self.changed_choices:
            table.add_row(
                "Choice symbols with different default selection between sdkconfig and Kconfig", style=SUBTITLE_STYLE
            )
            for choice_name, kconfig_selection, sdkconfig_selection in self.changed_choices:
                table.add_row(
                    f"{choice_name}: Kconfig default selection: {kconfig_selection}, "
                    f"sdkconfig default selection: {sdkconfig_selection}"
                )
            table.add_row("")

        if verbosity == VERBOSITY_VERBOSE and self.changed_values_promptless:
            table.add_row(
                "Invisible (promptless) config symbols with different values between sdkconfig and Kconfig",
                style="bold",
            )
            for sym_name, kconfig_value, sdkconfig_value, sdkconfig_value_is_default in self.changed_values_promptless:
                table.add_row(
                    (
                        f"{sym_name}: Kconfig default value: {kconfig_value}, "
                        f"sdkconfig value {sdkconfig_value} "
                        f"{'(user-set)' if sdkconfig_value_is_default else '(default)'} "
                        "will be ignored"
                    )
                )
            table.add_row("")

        return table

    def return_json(self) -> Optional[dict]:
        """
        Return the area report in JSON format.
        """
        if self._nothing_to_report():
            return None
        ret_json = super().return_json()
        if not ret_json:
            ret_json = dict()
        ret_json["data"] = dict()

        if self.changed_defaults:
            ret_json["data"]["changed_defaults"] = list()
            for sym_name, kconfig_value, sdkconfig_value in self.changed_defaults:
                ret_json["data"]["changed_defaults"].append(
                    {"name": sym_name, "kconfig_default": kconfig_value, "sdkconfig_default": sdkconfig_value}
                )
        if self.changed_values_promptless:  # There is all the info in json every time
            ret_json["data"]["mismatched_promptless"] = list()
            for sym_name, kconfig_value, sdkconfig_value, sdkconfig_value_is_default in self.changed_values_promptless:
                ret_json["data"]["mismatched_promptless"].append(
                    {
                        "name": sym_name,
                        "kconfig_value": kconfig_value,
                        "sdkconfig_value": sdkconfig_value,
                        "sdkconfig_value_is_default": sdkconfig_value_is_default,
                    }
                )
        if self.changed_choices:
            ret_json["data"]["changed_choices"] = list()
            for choice_name, kconfig_selection, sdkconfig_selection in self.changed_choices:
                ret_json["data"]["changed_choices"].append(
                    {
                        "name": choice_name,
                        "kconfig_selection": kconfig_selection,
                        "sdkconfig_selection": sdkconfig_selection,
                    }
                )

        return ret_json

    def reset(self) -> None:
        """
        Reset the area to its initial state, clearing all records and data.
        """
        self.changed_defaults.clear()
        self.changed_values_promptless.clear()
        self.changed_choices.clear()


class MultipleDefinitionArea(Area):
    """
    Multiple definition: having two or more definitions of the Symbol/Choice with the same name.
    NOTE: Currently, MultipleDefinitionArea cause only info instead of warning. This will be changed in the future.
    """

    def __init__(self):
        super().__init__(
            title="Multiple Symbol/Choice Definitions",
            ignore_codes=("multiple-definition", "MD"),
            info_string=textwrap.dedent(
                """\
                Multiple definitions of the same Symbol name are allowed by the Kconfig syntax.
                However, it may happen that e.g. two different components accidentally define the same Symbol name,
                which may lead to unexpected behavior.
                """
            ),
        )

        self.multiple_definitions: Dict[str, Set[str]] = dict()

    def add_record(self, sym_or_choice: "Union[Symbol, Choice]", **kwargs: Optional[dict]) -> None:
        """
        kwargs:
            occurrences: Set[str]
        """
        occurrences: Set[str] = kwargs["occurrences"] if "occurrences" in kwargs else set()  # type: ignore
        # For backward compatibility, code and type annotations allow unnamed choices
        sym_or_choice_name = sym_or_choice.name or "unnamed choice"
        if sym_or_choice_name not in self.multiple_definitions:
            self.multiple_definitions[sym_or_choice_name] = occurrences
        self.multiple_definitions[sym_or_choice_name].update(occurrences)

    def add_ignore(self, sym_or_choice):
        if sym_or_choice.__class__.__name__ == "Symbol":
            self.ignored_sc["configs"].add(sym_or_choice.name)
        else:
            self.ignored_sc["choices"].add(sym_or_choice.name)

    def report_severity(self) -> int:
        return STATUS_OK if not self.multiple_definitions else STATUS_OK_WITH_INFO

    def print(self, verbosity: str) -> Optional[Table]:
        if not self.multiple_definitions:
            return None

        table = Table(title=self.title, title_justify="left", show_header=False, title_style=AREA_TITLE_STYLE)
        table.box = HORIZONTALS
        table.add_column(
            "",
            justify="left",
            no_wrap=False,
        )
        table.add_row(
            "Multiple definitions will have higher severity in the future. Please, visit "
            "https://docs.espressif.com/projects/esp-idf-kconfig/en/latest/kconfiglib/configuration-report.html#multiple-symbol-or-choice-definition "  # noqa: E501
            "for more information.",
            style=INFO_STRING_STYLE,
        )
        if verbosity == VERBOSITY_VERBOSE:
            table.add_row(self.info_string, style=INFO_STRING_STYLE)

        for sym_or_choice_name in self.multiple_definitions:
            table.add_row(sym_or_choice_name, style="bold")
            for definition in self.multiple_definitions[sym_or_choice_name]:
                table.add_row(_INDENT + definition)

        return table

    def return_json(self) -> Optional[dict]:
        """
        Return the area report in JSON format.
        """
        if not self.multiple_definitions:
            return None
        ret_json = super().return_json()
        if not ret_json:
            ret_json = dict()
        ret_json["data"] = dict()
        for sym_or_choice_name in self.multiple_definitions:
            ret_json["data"][sym_or_choice_name] = list()
            for definition in self.multiple_definitions[sym_or_choice_name]:
                ret_json["data"][sym_or_choice_name].append(definition)
        return ret_json

    def reset(self) -> None:
        """
        Reset the area to its initial state, clearing all records and data.
        """
        self.multiple_definitions.clear()


class MiscArea(Area):
    """
    All the messages not related to the other areas.
    """

    def __init__(self):
        super().__init__(
            title="Miscellaneous",
            ignore_codes=set(),
            info_string="",
        )

        self.messages: Set[str] = set()

    def add_record(self, sym_or_choice: "Union[Symbol, Choice]", **kwargs: Optional[dict]) -> None:
        """
        kwargs:
            message: str
        As the only Area, MiscArea does not care about the Symbol/Choice.
        """
        if "message" not in kwargs.keys():
            raise AttributeError("Message must be specified for MiscArea.")
        self.messages.add(str(kwargs["message"]))

    def report_severity(self) -> int:
        return STATUS_OK if not self.messages else STATUS_OK_WITH_INFO

    def print(self, verbosity: str) -> Optional[Table]:
        if not self.messages:
            return None

        table = Table(title=self.title, title_justify="left", show_header=False, title_style=AREA_TITLE_STYLE)
        table.box = HORIZONTALS
        table.add_column("", justify="left", no_wrap=False)
        for message in self.messages:
            table.add_row(f"* {message}")
        return table

    def return_json(self) -> Optional[dict]:
        if not self.messages:
            return None
        ret_json = super().return_json()
        if not ret_json:
            ret_json = dict()
        ret_json["data"] = list(self.messages)
        return ret_json

    def reset(self) -> None:
        """
        Reset the area to its initial state, clearing all records and data.
        """
        self.messages.clear()


class MultipleAssignmentArea(Area):
    """
    This area reports multiple assignments to the same symbol within a single file.
    """

    def __init__(self):
        super().__init__(
            title="Multiple Assignments",
            ignore_codes=set(),
            info_string=(
                "Under normal circumstances, there should be only one (or none) assignment per config option. "
                "Multiple assignments mean somebody manually altered the sdkconfig file. "
                "If you edited the sdkconfig file intentionally, this area can be ignored."
            ),
        )

        # CONFIG_NAME: [(val, is_default?), (val, is_default?), ...]
        self.multiple_assignments_sym: Dict[Symbol, List[Tuple[str, bool]]] = defaultdict(list)
        # CHOICE_NAME: [(val, is_default?), (val, is_default?), ...]
        self.multiple_assignments_choice: Dict[Choice, List[Tuple[str, bool]]] = defaultdict(list)

    def add_record(self, sym_or_choice, **kwargs):
        """
        kwargs:
            new_value: str
            is_default: bool
        """
        if "new_value" not in kwargs.keys() or "is_default" not in kwargs.keys():
            raise AttributeError("New value and is_default must be specified for MultipleAssignmentArea.")
        if sym_or_choice.__class__.__name__ == "Symbol":
            if not self.multiple_assignments_sym[sym_or_choice]:
                # If this is the first time we are logging the data, we also need to log the first value
                self.multiple_assignments_sym[sym_or_choice] = [
                    (sym_or_choice.str_value, sym_or_choice._user_value is None)
                ]
            self.multiple_assignments_sym[sym_or_choice].append((kwargs["new_value"], kwargs["is_default"]))
        elif sym_or_choice.__class__.__name__ == "Choice":
            if not self.multiple_assignments_choice[sym_or_choice]:
                # If this is the first time we are logging the data, we also need to log the first selection
                self.multiple_assignments_choice[sym_or_choice] = [
                    (
                        sym_or_choice.selection.name if sym_or_choice.selection else "choice deselected",
                        sym_or_choice._user_selection is None,
                    )
                ]
            self.multiple_assignments_choice[sym_or_choice].append((kwargs["new_value"], kwargs["is_default"]))

    def report_severity(self) -> int:
        return STATUS_OK_WITH_INFO if (self.multiple_assignments_sym or self.multiple_assignments_choice) else STATUS_OK

    def print(self, verbosity: str) -> Optional[Table]:
        if self.report_severity() is STATUS_OK:
            return None

        table = Table(title=self.title, title_justify="left", show_header=False, title_style=AREA_TITLE_STYLE)
        table.box = HORIZONTALS
        table.add_column("", justify="left", no_wrap=False)

        if verbosity == VERBOSITY_VERBOSE:
            table.add_row(self.info_string, style=INFO_STRING_STYLE)

        for sym in self.multiple_assignments_sym:
            msg = f"Symbol {sym.name}: "
            for val, is_default in self.multiple_assignments_sym[sym]:
                msg += f"\n{_INDENT}{val}{' (default value)' if is_default else ' (user-set value)'}"
            msg += f"\n{_INDENT}-> using {sym.str_value}"
            table.add_row(msg)
        for choice in self.multiple_assignments_choice:
            msg = f"Choice {choice.name}: "
            for val, is_default in self.multiple_assignments_choice[choice]:
                msg += f"\n{_INDENT}{val}{' (default selection)' if is_default else ' (user-set selection)'}"
            msg += f"\n{_INDENT}-> using {choice.selection.name}"
            table.add_row(msg)

        return table

    def return_json(self) -> Optional[dict]:
        if not self.multiple_assignments_sym and not self.multiple_assignments_choice:
            return None
        ret_json = super().return_json()
        if not ret_json:
            ret_json = dict()
        ret_json["data"] = dict()
        ret_json["data"]["symbols"] = dict()
        ret_json["data"]["choices"] = defaultdict(list)

        for sym in self.multiple_assignments_sym:
            ret_json["data"]["symbols"][sym.name] = {"values": dict(), "final_value": sym.str_value}
            for val, is_default in self.multiple_assignments_sym[sym]:
                ret_json["data"]["symbols"][sym.name]["values"][val] = "default" if is_default else "user-set"

        for choice in self.multiple_assignments_choice:
            ret_json["data"]["choices"][choice.name] = {"values": dict(), "final_value": choice.selection.name}
            for val, is_default in self.multiple_assignments_choice[choice]:
                ret_json["data"]["choices"][choice.name]["values"][val] = "default" if is_default else "user-set"

        return ret_json

    def reset(self) -> None:
        """
        Reset the area to its initial state, clearing all records and data.
        """
        self.multiple_assignments_sym.clear()
        self.multiple_assignments_choice.clear()


class DisabledSymbolArea(Area):
    """
    This area reports symbols that are set via sdkconfig.defaults, but are not visible.
    """

    def __init__(self):
        super().__init__(
            title="Disabled Symbols/Choices With User-Set Value",
            ignore_codes=set(),
            info_string=(
                "Disabled symbols/choices are not written into sdkconfig file. Setting their user value has no effect."
            ),
        )
        self.hidden_symbols: Dict[Symbol, str] = dict()
        self.hidden_choices: Dict[Choice, Symbol] = dict()

    def add_record(self, sym_or_choice: "Union[Symbol, Choice]", **kwargs: Optional[dict]) -> None:
        """
        kwargs:
            user_value: str
        """
        if "user_value" not in kwargs.keys():
            raise AttributeError("User value must be specified for HiddenSymbolArea.")
        # Choice symbols are ignored; choice will be reported separately.
        if sym_or_choice.__class__.__name__ == "Symbol" and not sym_or_choice.choice:  # type: ignore
            self.hidden_symbols[sym_or_choice] = kwargs["user_value"]  # type: ignore
        elif sym_or_choice.__class__.__name__ == "Choice":
            self.hidden_choices[sym_or_choice] = kwargs["user_value"]  # type: ignore

    def report_severity(self) -> int:
        return STATUS_OK_WITH_INFO if self.hidden_symbols else STATUS_OK

    def print(self, verbosity: str) -> Optional[Table]:
        if not self.hidden_symbols:
            return None
        table = Table(title=self.title, title_justify="left", show_header=False, title_style=AREA_TITLE_STYLE)
        table.box = HORIZONTALS
        table.add_column("", justify="left", no_wrap=False, min_width=50)
        for symbol in self.hidden_symbols:
            table.add_row(
                f"* {symbol.name_and_loc} with user-set value {self.hidden_symbols[symbol]} from {symbol._user_source}"
            )
        for choice in self.hidden_choices:
            user_source = choice._user_source or getattr(choice._user_selection, "_user_source", "")
            table.add_row(
                f"* {choice.name_and_loc} with user-set symbol {self.hidden_choices[choice].name} from {user_source}"
            )
        return table

    def return_json(self) -> Optional[dict]:
        if not self.hidden_symbols:
            return None
        ret_json = super().return_json()
        if not ret_json:
            ret_json = dict()
        ret_json["data"] = dict()
        ret_json["data"]["symbols"] = dict()
        ret_json["data"]["choices"] = dict()
        for symbol in self.hidden_symbols:
            ret_json["data"]["symbols"][symbol.name] = {
                "value": self.hidden_symbols[symbol],
                "source": symbol._user_source,
            }
        for choice in self.hidden_choices:
            ret_json["data"]["choices"][choice.name] = {
                "value": self.hidden_choices[choice].name,
                "source": choice._user_source or getattr(choice._user_selection, "_user_source", ""),
            }

        return ret_json

    def reset(self) -> None:
        """
        Reset the area to its initial state, clearing all records and data.
        """
        self.hidden_symbols.clear()
        self.hidden_choices.clear()


class KconfigReport:
    """
    By add_record() method, new records are added to the report.
    Every time, it is needed to specify report area for given record.
    Every area is described by a class inheriting from Area class.

    NOTE: Full support will come in the future. Currently, only multiple-definition and misc areas are supported.
    """

    _instance = None
    _initialized: bool

    def __new__(cls, kconfig: "Kconfig", defaults_policy: DefaultsPolicy) -> "KconfigReport":
        """Singleton class to log messages"""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.defaults_policy = defaults_policy
            cls._instance._initialized = False
        cls._instance.kconfig = kconfig
        return cls._instance

    def __init__(
        self,
        kconfig: "Kconfig",
        defaults_policy: DefaultsPolicy,
    ) -> None:
        if hasattr(self, "_initialized") and self._initialized:
            return
        self._initialized = True

        self.kconfig: "Kconfig" = kconfig
        self.verbosity: str = os.getenv("KCONFIG_REPORT_VERBOSITY", VERBOSITY_DEFAULT)
        self.defaults_policy: DefaultsPolicy = defaults_policy

        # Ignores
        self.lines_with_ignores: List[str] = list()

        # Areas
        self.areas = (
            MultipleDefinitionArea(),
            MiscArea(),
            DefaultValuesArea(defaults_policy, verbosity=self.verbosity),
            MultipleAssignmentArea(),
            DisabledSymbolArea(),
        )

        # Mapping dictionaries
        """
        ignore_code_to_area:
            Mapping from ignore code to the Area object. It is used to quickly find the area for given ignore code.
        """
        self.ignore_code_to_area: Dict[str, Area] = {}
        for area in self.areas:
            for code in area.ignore_codes:
                self.ignore_code_to_area[code] = area

        """
        area_to_instance:
            Mapping from Area class to the Area object. It is used to quickly find the Area object for given Area class.
        """
        self.area_to_instance: Dict[type, Area] = {area.__class__: area for area in self.areas}

        """
        status (property):
            Status of the configuration report. It is calculated only once right before report is printed.
        """
        self._status: int = STATUS_NONE

    @property
    def status(self) -> int:
        """
        Get the status of the Configuration.
        """
        self._status = max(area.report_severity() for area in self.areas) or STATUS_OK

        return self._status

    def reset(self) -> None:
        """
        Reset the report to its initial state.
        """
        self.lines_with_ignores.clear()
        for area in self.areas:
            area.reset()
        self._status = STATUS_NONE

    def add_ignore_line(self, line: str) -> None:
        """
        Add line with ignore directive to ignored lines, which will be processed later.
        """
        if not self.kconfig:
            raise AttributeError("Kconfig object must be set before adding ignore lines.")
        parts = [part.strip() for part in line.split()]
        if parts[0] not in ["config", "choice", "menuconfig"]:
            self.add_record(
                MiscArea, message=f"Kconfig entry {parts[0]} does not support ignore directives. Directive ignored."
            )
            return
        sc_name = parts[1]
        for ignore_code in parts[4:]:  # config/choice NAME # ignore: CODE [CODE]* -> CODE is parts[4]
            try:
                self.ignore_code_to_area[ignore_code].add_ignore(self.kconfig._lookup_sym(sc_name))
            except KeyError:
                self.add_record(
                    MiscArea,
                    message=f'Ignoring ignore code "{ignore_code}" for {parts[0]} {sc_name} (unsupported ignore code).',
                )
                continue

    def add_record(self, area, **kwargs):
        """
        Adds a record for given area.
        Typically, record is related to the Symbol/Choice object (sym_or_choice from kwargs).
        However, some areas may not need sym_or_choice to be specified, thus it is not mandatory.
        """
        if "sym_or_choice" not in kwargs.keys():
            # E.g. MiscArea does not care about the Symbol/Choice
            self.area_to_instance[area].add_record(sym_or_choice=None, **kwargs)
        else:
            self.area_to_instance[area].add_record(**kwargs)

    def _make_header(self) -> Table:
        if not self.kconfig:
            raise AttributeError("Kconfig object must be set before printing report.")
        header_table = Table(title_style="bold", show_header=False)
        header_table.box = None
        header_table.add_column("Configuration", justify="left")
        header_table.add_row(f"Parser Version: {self.kconfig.parser_version}")
        header_table.add_row(f"Verbosity: {self.verbosity}")
        if self.verbosity == VERBOSITY_VERBOSE:
            header_table.add_row(f"Symbols parsed: {len(self.kconfig.unique_defined_syms)}")

        header_table.add_row(f"Defaults policy: {self.defaults_policy.value}")
        status = self.status
        if status == STATUS_OK:
            header_table.add_row("Status: Finished successfully", style="green")
        elif status == STATUS_OK_WITH_INFO:
            header_table.add_row("Status: Finished with notifications", style="green_yellow")
            if self.verbosity == VERBOSITY_VERBOSE:
                header_table.add_row(
                    "Configuration is successfully finished, but the system has identified situations that could "
                    "potentially cause some issues. Please check the relevant areas.",
                    style="green_yellow",
                )
        elif status == STATUS_WARNING:
            header_table.add_row("Status: Finished with warnings", style="yellow")
            if self.verbosity == VERBOSITY_VERBOSE:
                header_table.add_row(
                    "Configuration is successfully finished, but the system has identified situations that probably "
                    "will cause some issues. Please check the relevant areas.",
                    style="yellow",
                )
        else:  # NOTE: STATUS_ERROR is currently not used, so "failed" here is just a placeholder.
            header_table.add_row("Status: Failed", style="red")

        header_table.add_row("")

        return header_table

    def _finalize_report_data(self) -> None:
        """
        Gather information available only after Kconfig files are parsed and sdkconfig files are loaded.
        NOTE: If possible, log the information on the fly. This method serves only for areas that
              cannot be logged that way.
        """
        # we cannot use BOOL_TO_STR here because it would lead to a circular import
        bool_to_str = {0: "n", 2: "y"}
        # Print a notification if symbol/choice has a user value, but is not visible.
        for sym in self.kconfig.defined_syms:
            if sym._user_value is not None and sym.visibility == 0:
                str_value = bool_to_str[sym._user_value] if sym._user_value in bool_to_str else sym._user_value
                self.add_record(DisabledSymbolArea, sym_or_choice=sym, user_value=str_value)
        for choice in self.kconfig.unique_choices:
            if choice._user_selection is not None and choice.visibility == 0:
                self.add_record(DisabledSymbolArea, sym_or_choice=choice, user_value=choice._user_selection)

    def print_report(self, file: Optional[str] = None) -> None:
        # Some areas (DisabledSymbolArea for instance) need to be finalized after Kconfig
        # files are parsed and sdkconfig files are loaded.
        self._finalize_report_data()

        if self.verbosity == VERBOSITY_QUIET and self.status in (STATUS_OK, STATUS_OK_WITH_INFO, STATUS_WARNING):
            return

        report_table = Table(title="Configuration Report", title_style="bold", show_header=False, title_justify="left")
        report_table.box = HORIZONTALS
        report_table.add_column("Configuration", justify="center", no_wrap=False)
        report_table.add_row(self._make_header())

        for area in self.areas:
            sub_report = area.print(verbosity=self.verbosity)
            if sub_report:
                report_table.add_row(sub_report)

        if not file:
            console = Console(force_terminal=True, stderr=True)
            console.print(report_table)
        else:
            with open(file, "w") as f:
                rprint(report_table, file=f)

    def output_json(self, file: Optional[str] = None) -> None:
        report_json = self._return_json()

        if not file:
            console = Console(force_terminal=True, stderr=True)
            console.print(json.dumps(report_json, indent=4))
        else:
            with open(file, "w+") as f:
                json.dump(report_json, f, indent=4)

    def _return_json(self) -> Dict:
        """
        Return the report in JSON format.
        """
        self._finalize_report_data()
        report_json: Dict = dict()
        report_json["header"] = dict()
        report_json["header"]["report_type"] = "kconfig"
        report_json["header"]["parser_version"] = self.kconfig.parser_version
        report_json["header"]["verbosity"] = self.verbosity
        report_json["header"]["status"] = Area.severity_to_str(self.status)
        report_json["header"]["unique_defined_syms"] = len(self.kconfig.unique_defined_syms)
        report_json["header"]["defaults_policy"] = self.defaults_policy.value

        report_json["areas"] = list()
        for area in self.areas:
            # Status OK means that there is nothing to report in the area
            if area.report_severity() == STATUS_OK:
                continue
            report_json["areas"].append(area.return_json())

        return report_json



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class DefaultsPolicy(Enum):
    USE_SDKCONFIG = "sdkconfig"
    INTERACTIVE = "interactive"
    USE_KCONFIG = "kconfig"

    @property
    def description(self) -> str:
        """Return a human-readable description of the policy."""
        if self == DefaultsPolicy.USE_SDKCONFIG:
            return "Using sdkconfig default values"
        elif self == DefaultsPolicy.INTERACTIVE:
            return "Default value mismatches resolved interactively by the user"
        elif self == DefaultsPolicy.USE_KCONFIG:
            return "Using Kconfig default values"
        else:
            return "Unknown policy"


# Start and end comment for deprecated options written into sdkconfig
DEP_OP_BEGIN = "# Deprecated options for backward compatibility"
DEP_OP_END = "# End of deprecated options"

# This pragma marks config options that have default value in sdkconfig file:
SDKCONFIG_DEFAULT_PRAGMA = "# default:"



#!/usr/bin/env python
#
# Command line tool to check kconfig files
#
# SPDX-FileCopyrightText: 2018-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import argparse
import os
import re
from typing import Optional
from typing import Tuple

from .check_deprecated_options import _prepare_deprecated_options
from .check_deprecated_options import check_deprecated_options

# output file with suggestions will get this suffix
OUTPUT_SUFFIX = ".new"

CONFIG_PREFIX = "CONFIG_"

SPACES_PER_INDENT = 4

CONFIG_NAME_MAX_LENGTH = 50

CONFIG_NAME_MIN_PREFIX_LENGTH = 3

# The checker will not fail if it encounters this string (it can be used for temporarily resolve conflicts)
RE_NOERROR = re.compile(r"\s+#\s+NOERROR\s+$")

# list or rules for lines
LINE_ERROR_RULES = [
    # (regular expression for finding,      error message,                                  correction)
    (
        re.compile(r"\t"),
        "tabulators should be replaced by spaces",
        r" " * SPACES_PER_INDENT,
    ),
    (re.compile(r"\s+\n"), "trailing whitespaces should be removed", r"\n"),
    (re.compile(r".{120}"), "line should be shorter than 120 characters", None),
]


class InputError(RuntimeError):
    """
    Represents and error on the input
    """

    def __init__(self, path, line_number, error_msg, suggested_line):
        super(InputError, self).__init__("{}:{}: {}".format(path, line_number, error_msg))
        self.suggested_line = suggested_line


class BaseChecker(object):
    """
    Base class for all checker objects
    """

    def __init__(self, path_in_idf):
        self.path_in_idf = path_in_idf

    def finalize(self):
        """
        Abstract method for finalizing the checker
        """
        pass

    def process_line(self, line, line_number):
        """
        Abstract method for processing a line
        """
        raise NotImplementedError("process_line method is not implemented in BaseChecker")


class SourceChecker(BaseChecker):
    # allow to source only files which will be also checked by the script
    # Note: The rules are complex and the LineRuleChecker cannot be used
    def process_line(self, line, line_number):
        m = re.search(r'^\s*[ro]{0,2}source(\s*)"([^"]+)"', line)

        if m:
            if len(m.group(1)) == 0:
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    '"source" has to been followed by space',
                    line.replace("source", "source "),
                )
            path = m.group(2)
            filename = os.path.basename(path)
            allowed_envvars = [
                "$COMPONENT_KCONFIGS_SOURCE_FILE",
                "$COMPONENT_KCONFIGS_PROJBUILD_SOURCE_FILE",
                "$COMPONENT_KCONFIGS_EXCLUDED_SOURCE_FILE",
                "$COMPONENT_KCONFIGS_PROJBUILD_EXCLUDED_SOURCE_FILE",
            ]
            # skip the rest of the checks for variables that do not need to explicitly specify file name
            if path in allowed_envvars:
                return
            # path == filename means path is only an env_var, without file name explicitly mentioned
            if path.startswith("$") and path == filename:
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    (
                        "environment variables cannot specify the filename in sourced files, only the path. "
                        'Specify the filename explicitly ("$ENV_VAR/Kconfig.<suffix>").'
                    ),
                    line.replace(path, os.path.join(path, "Kconfig.<suffix>")),
                )
            elif not filename.startswith("Kconfig."):
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    "only filenames starting with Kconfig.* can be sourced",
                    line.replace(path, os.path.join(os.path.dirname(path), "Kconfig." + filename)),
                )


class LineRuleChecker(BaseChecker):
    """
    checks LINE_ERROR_RULES for each line
    """

    def process_line(self, line, line_number):
        suppress_errors = RE_NOERROR.search(line) is not None
        errors = []
        for rule in LINE_ERROR_RULES:
            m = rule[0].search(line)
            if m:
                if suppress_errors:
                    # just print but no failure
                    e = InputError(self.path_in_idf, line_number, rule[1], line)
                    print(f"NOERROR: {e}")
                else:
                    errors.append(rule[1])
                if rule[2]:
                    line = rule[0].sub(rule[2], line)
        if len(errors) > 0:
            raise InputError(self.path_in_idf, line_number, "; ".join(errors), line)


class ConfigNameChecker(BaseChecker):
    """
    Checks if the global syntax of config/symbol names is correct.
    Every rule is in separate method to allow easy extension.
    """

    def __init__(self, path_in_idf):
        super().__init__(path_in_idf)

    def rule_prefix(self, config_name: str, line: str, line_number: int) -> None:
        if not config_name.startswith(CONFIG_PREFIX):
            config_name_start = line.find(config_name)
            raise InputError(
                self.path_in_idf,
                line_number,
                f"{config_name} should start with {CONFIG_PREFIX}",
                f"{line[:config_name_start]}{CONFIG_PREFIX}{line[config_name_start:]}",
            )

    def rule_uppercase(self, config_name: str, line: str, line_number: int) -> None:
        if not config_name.isupper():
            raise InputError(
                self.path_in_idf,
                line_number,
                f"{config_name} should be all uppercase",
                line.replace(config_name, config_name.upper()),
            )

    def rule_name_len(self, config_name: str, line: str, line_number: int) -> None:
        name_length = len(config_name)
        # When checking Kconfig files, there is not yet a prefix and thus, 50-char length is checked without it.
        # To be consistent, prefix is not counted in the length.
        if name_length - len(CONFIG_PREFIX) > CONFIG_NAME_MAX_LENGTH:
            raise InputError(
                self.path_in_idf,
                line_number,
                f"{config_name} is {name_length} characters long and it should be {CONFIG_NAME_MAX_LENGTH} at most",
                line,
            )  # no suggested correction for this

    def process_line(self, line: str, line_number: int) -> None:
        raise NotImplementedError("ConfigNameChecker checks config names, not lines.")

    def process_config_name(self, config_name: str, line: str, line_number: int) -> None:
        for rule in (self.rule_prefix, self.rule_uppercase, self.rule_name_len):
            rule(config_name, line, line_number)


class IndentAndNameChecker(BaseChecker):
    """
    checks the indentation of each line and configuration names
    """

    def __init__(self, path_in_idf, debug=False):
        super(IndentAndNameChecker, self).__init__(path_in_idf)
        self.debug = debug
        self.min_prefix_length = CONFIG_NAME_MIN_PREFIX_LENGTH

        # stack of the nested menuconfig items, e.g. ['mainmenu', 'menu', 'config']
        self.level_stack = []

        # stack common prefixes of configs
        self.prefix_stack = []

        # if the line ends with '\' then we force the indent of the next line
        self.force_next_indent = 0

        # menu items which increase the indentation of the next line
        self.re_increase_level = re.compile(
            r"""^\s*
                                          (
                                               (menu(?!config))
                                              |(mainmenu)
                                              |(choice)
                                              |(config)
                                              |(menuconfig)
                                              |(help)
                                              |(if)
                                              |(source)
                                              |(osource)
                                              |(rsource)
                                              |(orsource)
                                          )
                                       """,
            re.X,
        )

        # closing menu items which decrease the indentation
        self.re_decrease_level = re.compile(
            r"""^\s*
                                          (
                                               (endmenu)
                                              |(endchoice)
                                              |(endif)
                                          )
                                       """,
            re.X,
        )

        # matching beginning of the closing menuitems
        self.pair_dic = {
            "endmenu": "menu",
            "endchoice": "choice",
            "endif": "if",
        }

        # regex for config names
        self.re_name = re.compile(
            r"""^
                                       (
                                            (?:config)
                                           |(?:menuconfig)
                                           |(?:choice)

                                       )\s+
                                       (\w+)
                                      """,
            re.X,
        )

        # regex for new prefix stack
        self.re_new_stack = re.compile(
            r"""^
                                            (
                                                 (?:menu(?!config))
                                                |(?:mainmenu)
                                                |(?:choice)

                                            )
                                            """,
            re.X,
        )

        # regexes to get lines containing expressions
        # Unquoted symbols are either config names, y/n or (hex)num literals.
        # Catching also no-uppercase config names (TyPO_NAME) to throw an error later on.
        # Quoted symbols are "y"/"n", env_vars or string literals;
        # the last two categories can contain anything between the quotes, thus it is broader.
        symbol = r"\w+|\".+?\"|'.+?'"
        reg_prompt = re.compile(r"^\".*?\"\s+(?:if)\s+(?P<expression0>.*)$")
        reg_default = re.compile(r"^(?P<expression0>.*)\s+(?:if)\s+(?P<expression1>.*)$")
        reg_select_imply = re.compile(rf"^(?P<expression0>{symbol})\s+(?:if)\s+(?P<expression1>.*)$")
        reg_range = re.compile(
            rf"^(?P<expression0>{symbol})\s+(?P<expression1>{symbol})\s+(?:if)\s+(?P<expression2>.*)$"
        )
        reg_depends_on_visible_if = re.compile(r"^(?P<expression0>.*)$")
        reg_config_menuconfig_choice = re.compile(rf"^(?P<expression>{symbol})$")
        self.reg_switch = re.compile(
            r"^\s*(?P<keyword>prompt|default|select|imply|range|depends on|config|menuconfig|choice|visible if)\s+(?P<body>.*)$"  # noqa: E501
        )
        self.reg_symbol = re.compile(rf"{symbol}", re.X)

        self.kw_to_regex = {
            "prompt": reg_prompt,
            "default": reg_default,
            "select": reg_select_imply,
            "imply": reg_select_imply,
            "range": reg_range,
            "depends on": reg_depends_on_visible_if,
            "visible if": reg_depends_on_visible_if,
            "config": reg_config_menuconfig_choice,
            "menuconfig": reg_config_menuconfig_choice,
            "choice": reg_config_menuconfig_choice,
        }

    def finalize(self):
        if len(self.prefix_stack) > 0:
            self.check_common_prefix("", "EOF")
        if len(self.prefix_stack) != 0:
            if self.debug:
                print(self.prefix_stack)
            raise RuntimeError("Prefix stack should be empty. Perhaps a menu/choice hasn't been closed")

    def del_from_level_stack(self, count):
        """delete count items from the end of the level_stack"""
        if count > 0:
            # del self.level_stack[-0:] would delete everything and we expect not to delete anything for count=0
            del self.level_stack[-count:]

    def update_level_for_inc_pattern(self, new_item):
        if self.debug:
            print("level+", new_item, ": ", self.level_stack, end=" -> ")
        # "config" and "menuconfig" don't have a closing pair. So if new_item is an item which need to be indented
        # outside the last "config" or "menuconfig" then we need to find to a parent where it belongs
        if new_item in [
            "config",
            "menuconfig",
            "menu",
            "choice",
            "if",
            "source",
            "rsource",
            "osource",
            "orsource",
        ]:
            # item is not belonging to a previous "config" or "menuconfig" so need to indent to parent
            for i, item in enumerate(reversed(self.level_stack)):
                if item in ["menu", "mainmenu", "choice", "if"]:
                    # delete items ("config", "menuconfig", "help") until the appropriate parent
                    self.del_from_level_stack(i)
                    break
            else:
                # delete everything when configs are at top level without a parent menu, mainmenu...
                self.del_from_level_stack(len(self.level_stack))

        self.level_stack.append(new_item)
        if self.debug:
            print(self.level_stack)
        # The new indent is for the next line. Use the old one for the current line:
        return len(self.level_stack) - 1

    def update_level_for_dec_pattern(self, new_item):
        if self.debug:
            print("level-", new_item, ": ", self.level_stack, end=" -> ")
        target = self.pair_dic[new_item]
        for i, item in enumerate(reversed(self.level_stack)):
            # find the matching beginning for the closing item in reverse-order search
            # Note: "menuconfig", "config" and "help" don't have closing pairs and they are also on the stack. Now they
            # will be deleted together with the "menu" or "choice" we are closing.
            if item == target:
                i += 1  # delete also the matching beginning
                if self.debug:
                    print("delete ", i, end=" -> ")
                self.del_from_level_stack(i)
                break
        if self.debug:
            print(self.level_stack)
        return len(self.level_stack)

    def check_name_sanity(self, line: str, line_number: int) -> None:
        def is_hex(s: str) -> bool:
            return re.search(r"^0x[0-9a-fA-F]+$", s) is not None

        line = line[: line.index("#")] + "\n" if "#" in line else line
        line_with_symbols = self.reg_switch.match(line)

        if line_with_symbols:
            expressions = self.kw_to_regex[line_with_symbols.group("keyword")].match(line_with_symbols.group("body"))
            if expressions:
                for k in expressions.groupdict().keys():
                    symbols = self.reg_symbol.findall(expressions.groupdict()[k])
                    for symbol in symbols:
                        if not (
                            symbol in ("y", '"y"', "n", '"n"')
                            or symbol.isupper()
                            or symbol.isnumeric()
                            or is_hex(symbol)
                            or symbol.startswith(('"', "'"))
                        ):
                            raise InputError(
                                self.path_in_idf,
                                line_number,
                                f"config name {symbol} should be all uppercase",
                                line.replace(symbol, symbol.upper()),
                            )

    def check_name_and_update_prefix(self, line, line_number):
        m = self.re_name.search(line)
        if m:
            name = m.group(2)
            name_length = len(name)

            if name_length > CONFIG_NAME_MAX_LENGTH:
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    f"{name} is {name_length} characters long and it should be {CONFIG_NAME_MAX_LENGTH} at most",
                    line + "\n",
                )  # no suggested correction for this
            if len(self.prefix_stack) == 0:
                self.prefix_stack.append(name)
            elif self.prefix_stack[-1] is None:
                self.prefix_stack[-1] = name
            else:
                # this has nothing common with paths but the algorithm can be used for this also
                self.prefix_stack[-1] = os.path.commonprefix([self.prefix_stack[-1], name])
            if self.debug:
                print("prefix+", self.prefix_stack)
        m = self.re_new_stack.search(line)
        if m:
            self.prefix_stack.append(None)
            if self.debug:
                print("prefix+", self.prefix_stack)

    def check_common_prefix(self, line, line_number):
        common_prefix = self.prefix_stack.pop()
        if self.debug:
            print("prefix-", self.prefix_stack)
        if common_prefix is None:
            return
        common_prefix_len = len(common_prefix)
        if common_prefix_len < self.min_prefix_length:
            raise InputError(
                self.path_in_idf,
                line_number,
                'The common prefix for the config names of the menu ending at this line is "{}".\n'
                "\tAll config names in this menu should start with the same prefix of {} characters "
                "or more.".format(common_prefix, self.min_prefix_length),
                line,
            )  # no suggested correction for this
        if len(self.prefix_stack) > 0:
            parent_prefix = self.prefix_stack[-1]
            if parent_prefix is None:
                # propagate to parent level where it will influence the prefix checking with the rest which might
                # follow later on that level
                self.prefix_stack[-1] = common_prefix
            else:
                if len(self.level_stack) > 0 and self.level_stack[-1] in [
                    "mainmenu",
                    "menu",
                ]:
                    # the prefix from menu is not required to propagate to the children
                    return
                if not common_prefix.startswith(parent_prefix):
                    raise InputError(
                        self.path_in_idf,
                        line_number,
                        f"Common prefix '{common_prefix}' should start with {parent_prefix}",
                        line,
                    )  # no suggested correction for this

    def process_line(self, line, line_number):
        stripped_line = line.strip()
        if len(stripped_line) == 0:
            self.force_next_indent = 0
            return
        # Ignore comment lines
        if stripped_line.startswith("#"):
            return
        current_level = len(self.level_stack)
        m = re.search(r"\S", line)  # indent found as the first non-space character
        if m:
            current_indent = m.start()
        else:
            current_indent = 0

        if current_level > 0 and self.level_stack[-1] == "help":
            if current_indent >= current_level * SPACES_PER_INDENT:
                # this line belongs to 'help'
                self.force_next_indent = 0
                return

        if self.force_next_indent > 0:
            if current_indent != self.force_next_indent:
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    "Indentation consists of {} spaces instead of {}".format(current_indent, self.force_next_indent),
                    (" " * self.force_next_indent) + line.lstrip(),
                )
            else:
                if not stripped_line.endswith("\\"):
                    self.force_next_indent = 0
                return

        elif stripped_line.endswith("\\") and stripped_line.startswith(("config", "menuconfig", "choice")):
            raise InputError(
                self.path_in_idf,
                line_number,
                "Line-wrap with backslash is not supported here",
                line,
            )  # no suggestion for this

        m = self.re_increase_level.search(line)
        if m:
            current_level = self.update_level_for_inc_pattern(m.group(1))
        else:
            m = self.re_decrease_level.search(line)
            if m:
                new_item = m.group(1)
                current_level = self.update_level_for_dec_pattern(new_item)
                if new_item not in ["endif"]:
                    # endif doesn't require to check the prefix because the items inside if/endif belong to the
                    # same prefix level
                    self.check_common_prefix(line, line_number)

        # name has to be checked after increase/decrease indentation level
        # otherwise false-positive indentation error for lines below name is raised
        self.check_name_sanity(line, line_number)
        self.check_name_and_update_prefix(stripped_line, line_number)

        expected_indent = current_level * SPACES_PER_INDENT

        if stripped_line.endswith("\\"):
            self.force_next_indent = expected_indent + SPACES_PER_INDENT
        else:
            self.force_next_indent = 0

        if current_indent != expected_indent:
            raise InputError(
                self.path_in_idf,
                line_number,
                "Indentation consists of {} spaces instead of {}".format(current_indent, expected_indent),
                (" " * expected_indent) + line.lstrip(),
            )


class SDKRenameChecker(BaseChecker):
    """
    Checks sdkconfig.rename[.target] files:
    * if the line contains at least two tokens (old and new config name pairs)
    * if the inversion syntax is used correctly
    """

    def __init__(self, path_in_idf):
        super().__init__(path_in_idf)
        self.renames = dict()  # key = old name, value = new name
        self.config_name_checker = ConfigNameChecker(path_in_idf)

    def process_line(self, line: str, line_number: int) -> None:
        # The line should look like CONFIG_OLD_NAME CONFIG_NEW_NAME # optional comment,
        # just # comment or blank line.
        if line.startswith("#") or line.isspace():
            return

        tokens = line.split()
        if len(tokens) < 2:
            raise InputError(
                self.path_in_idf,
                line_number,
                "Line should contain at least old and new config names.",
                line,  # no suggestions for this
            )
        else:
            old_name, new_name = tokens[:2]  # [:2] removes optional comment from tokens

        inversion = False
        if new_name.startswith("!"):  # inversion
            new_name = new_name[1:]
            inversion = True

        ############################################
        # sdkconfig.rename specific checks
        ############################################
        # Check inversion syntax
        if old_name.startswith("!"):
            raise InputError(
                self.path_in_idf,
                line_number,
                "For inversion, use CONFIG_OLD !CONFIG_NEW syntax",
                line[: line.index(old_name)]
                + old_name[1:]
                + line[line.index(old_name) + len(old_name) : line.index(new_name)]
                + "!"
                + line[line.index(new_name) :],
            )

        # Check for duplicit lines
        if old_name in self.renames.keys() and self.renames[old_name] == new_name:
            raise InputError(
                self.path_in_idf,
                line_number,
                f"There is a duplicit line: {old_name} {new_name if not inversion else '!' + new_name}",
                "",  # omit the duplicate
            )

        # Check if the there is repeated rename from the old name
        # This is not allowed because if the two different new names would have different values,
        # the result would be ambiguous
        # NOTE: Multiple renames to the same new name are allowed.
        #       In that case, the value of the new name is just propagated to all the old names.
        if old_name in self.renames.keys():
            raise InputError(
                self.path_in_idf,
                line_number,
                f"There already is a renaming from an old config name {old_name}",
                line,  # no suggestions for this
            )

        ############################################
        # Checking correctness of the config names
        ############################################
        # We want to check only if correct prefix is used, old names may not comply with the rules
        self.config_name_checker.rule_prefix(old_name, line, line_number)
        self.config_name_checker.process_config_name(new_name, line, line_number)

        # Add the new renaming to the dictionary
        self.renames[old_name] = new_name


def valid_directory(path):
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError("{} is not a valid directory!".format(path))
    return path


def validate_file(file_full_path: str, verbose: bool = False, replace: bool = False) -> bool:
    # Even in case of in_place modification, create a new file with suggestions (original will be replaced later).
    suggestions_full_path = file_full_path + OUTPUT_SUFFIX
    fail = False
    checkers: Tuple[BaseChecker, ...] = tuple()
    if "Kconfig" in os.path.basename(file_full_path):
        checkers = (
            IndentAndNameChecker(file_full_path),  # indent checker has to be before line checker, otherwise
            # false-positive indent error if error in line_checker
            LineRuleChecker(file_full_path),
            SourceChecker(file_full_path),
        )
    elif "sdkconfig.rename" in os.path.basename(file_full_path):
        checkers = (SDKRenameChecker(file_full_path),)

    if not checkers:
        raise NotImplementedError(
            f"The {os.path.basename(file_full_path)} files are not currently supported by kconfcheck."
        )

    with open(file_full_path, "r", encoding="utf-8") as f, open(
        suggestions_full_path, "w", encoding="utf-8", newline="\n"
    ) as f_o:
        try:
            for line_number, line in enumerate(f):
                try:
                    for checker in checkers:
                        checker.process_line(line=line, line_number=line_number)
                    # The line is correct therefore we echo it to the output file
                    f_o.write(line)
                except InputError as e:
                    print(e)
                    fail = True
                    f_o.write(e.suggested_line)
        except UnicodeDecodeError:
            raise ValueError("The encoding of {} is not Unicode.".format(file_full_path))
        finally:
            for checker in checkers:
                try:
                    checker.finalize()
                except InputError as e:
                    print(e)
                    fail = True

    if replace:
        os.replace(suggestions_full_path, file_full_path)

    if fail:
        print(
            "\t{} has been saved with suggestions for resolving the issues.\n"
            "\tPlease note that the suggestions can be wrong and "
            "you might need to re-run the checker several times "
            "for solving all issues".format(suggestions_full_path if not replace else file_full_path)
        )
        return False
    else:
        print("{}: OK".format(file_full_path))
        # If replace, file already removed
        if not replace:
            try:
                os.remove(suggestions_full_path)
            except Exception:
                # It is not a serious error if the file cannot be deleted
                print(f"{suggestions_full_path} cannot be deleted!")
        return True


def main() -> int:
    parser = argparse.ArgumentParser(description="Kconfig style checker")
    parser.add_argument(
        "--check",
        "-c",
        choices=["syntax", "deprecated"],
        default="syntax",
        help="Check syntax or deprecated options",
    )
    parser.add_argument(
        "files",
        nargs="*",
        help="Kconfig files to check (full paths separated by space)",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Print more information (useful for debugging)",
    )
    parser.add_argument(
        "--replace",
        action="store_true",
        help="[only for --check syntax] Apply the changes to the original files"
        f" instead of creating {OUTPUT_SUFFIX} files",
    )
    parser.add_argument(
        "--includes",
        "-d",
        nargs="*",
        help="[only for --check deprecated] Paths for recursive search of sdkconfig files",
        type=valid_directory,
    )
    parser.add_argument(
        "--exclude-submodules",
        nargs="*",
        type=valid_directory,
        help="[only for --check deprecated] Exclude ESP-IDF submodules",
    )

    args = parser.parse_args()

    if args.check == "syntax" and (args.includes is not None or args.exclude_submodules is not None):
        raise argparse.ArgumentError(
            None, "--includes and --exclude-submodules are available only when using --check deprecated option."
        )

    if args.check == "deprecated" and args.replace:
        raise argparse.ArgumentError(None, "--replace is available only when using --check syntax option.")

    success_counter = 0
    failure_counter = 0
    ignored_counter = 0

    files = [os.path.abspath(file_path) for file_path in args.files]

    if args.check == "deprecated":
        files, deprecated_options, ignore_dirs = _prepare_deprecated_options(
            args.includes, args.exclude_submodules, files
        )

    for full_path in files:
        file_ok: Optional[bool] = False
        if args.check == "syntax":
            file_ok = validate_file(full_path, args.verbose, args.replace)
        elif args.check == "deprecated":
            file_ok = check_deprecated_options(full_path, deprecated_options, ignore_dirs)
        else:
            raise argparse.ArgumentError(None, f"Unknown check type: {args.check} passed to --check argument.")

        if file_ok is None:
            ignored_counter += 1
        elif file_ok is True:
            success_counter += 1
        else:
            failure_counter += 1

    def _handle_plural(cnt: int) -> str:
        return "s" if cnt > 1 else ""

    if success_counter > 0:
        print(f"{success_counter} file{_handle_plural(success_counter)} have been successfully checked.")
    if ignored_counter > 0:
        print(f"{ignored_counter} file{_handle_plural(success_counter)} have been ignored.")
    if failure_counter > 0:
        print(f"{failure_counter} file{_handle_plural(success_counter)} have errors. Please take a look at the log.")
        return 1

    if not files:
        print("WARNING: no files specified.")
    return 0
